#   Copyright 2015 Symantec Corporation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from unittest import mock

from openstack.compute.v2 import flavor as _flavor
from openstack import exceptions as sdk_exceptions
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.compute.v2 import flavor
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestFlavor(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.projects_mock = self.identity_client.projects
        self.projects_mock.reset_mock()


class TestFlavorCreate(TestFlavor):
    def setUp(self):
        super().setUp()

        self.flavor = sdk_fakes.generate_fake_resource(
            _flavor.Flavor, links='flavor-links'
        )
        self.project = sdk_fakes.generate_fake_resource(_project.Project)

        self.columns = (
            'OS-FLV-DISABLED:disabled',
            'OS-FLV-EXT-DATA:ephemeral',
            'description',
            'disk',
            'id',
            'name',
            'os-flavor-access:is_public',
            'properties',
            'ram',
            'rxtx_factor',
            'swap',
            'vcpus',
        )
        self.data = (
            self.flavor.is_disabled,
            self.flavor.ephemeral,
            self.flavor.description,
            self.flavor.disk,
            self.flavor.id,
            self.flavor.name,
            self.flavor.is_public,
            format_columns.DictColumn(self.flavor.extra_specs),
            self.flavor.ram,
            self.flavor.rxtx_factor,
            self.flavor.swap,
            self.flavor.vcpus,
        )
        self.data_private = (
            self.flavor.is_disabled,
            self.flavor.ephemeral,
            self.flavor.description,
            self.flavor.disk,
            self.flavor.id,
            self.flavor.name,
            False,
            format_columns.DictColumn(self.flavor.extra_specs),
            self.flavor.ram,
            self.flavor.rxtx_factor,
            self.flavor.swap,
            self.flavor.vcpus,
        )

        self.projects_mock.get.return_value = self.project
        self.compute_client.create_flavor.return_value = self.flavor

        self.cmd = flavor.CreateFlavor(self.app, None)

    def test_flavor_create_default_options(self):
        arglist = [self.flavor.name]
        verifylist = [
            ('name', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        default_args = {
            'name': self.flavor.name,
            'ram': 256,
            'vcpus': 1,
            'disk': 0,
            'id': None,
            'ephemeral': 0,
            'swap': 0,
            'is_public': True,
        }

        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_flavor.assert_called_once_with(
            **default_args
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_flavor_create_all_options(self):
        self.set_compute_api_version('2.55')

        arglist = [
            '--id',
            self.flavor.id,
            '--ram',
            str(self.flavor.ram),
            '--disk',
            str(self.flavor.disk),
            '--ephemeral',
            str(self.flavor.ephemeral),
            '--swap',
            str(self.flavor.swap),
            '--vcpus',
            str(self.flavor.vcpus),
            '--rxtx-factor',
            str(self.flavor.rxtx_factor),
            '--public',
            '--description',
            str(self.flavor.description),
            '--property',
            'property=value',
            self.flavor.name,
        ]
        verifylist = [
            ('id', self.flavor.id),
            ('ram', self.flavor.ram),
            ('disk', self.flavor.disk),
            ('ephemeral', self.flavor.ephemeral),
            ('swap', self.flavor.swap),
            ('vcpus', self.flavor.vcpus),
            ('rxtx_factor', self.flavor.rxtx_factor),
            ('public', True),
            ('description', self.flavor.description),
            ('properties', {'property': 'value'}),
            ('name', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        args = {
            'name': self.flavor.name,
            'ram': self.flavor.ram,
            'vcpus': self.flavor.vcpus,
            'disk': self.flavor.disk,
            'id': self.flavor.id,
            'ephemeral': self.flavor.ephemeral,
            'swap': self.flavor.swap,
            'rxtx_factor': self.flavor.rxtx_factor,
            'is_public': self.flavor.is_public,
            'description': self.flavor.description,
        }

        props = {'property': 'value'}

        # SDK updates the flavor object instance. In order to make the
        # verification clear and preciese let's create new flavor and change
        # expected props this way
        create_flavor = _flavor.Flavor(**self.flavor)
        expected_flavor = _flavor.Flavor(**self.flavor)
        expected_flavor.extra_specs = props
        # convert expected data tuple to list to be able to modify it
        cmp_data = list(self.data)
        cmp_data[7] = format_columns.DictColumn(props)
        self.compute_client.create_flavor.return_value = create_flavor
        self.compute_client.create_flavor_extra_specs.return_value = (
            expected_flavor
        )

        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_flavor.assert_called_once_with(**args)
        self.compute_client.create_flavor_extra_specs.assert_called_once_with(
            create_flavor, props
        )
        self.compute_client.get_flavor_access.assert_not_called()

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(tuple(cmp_data), data)

    def test_flavor_create_other_options(self):
        self.set_compute_api_version('2.55')

        self.flavor.is_public = False
        arglist = [
            '--id',
            'auto',
            '--ram',
            str(self.flavor.ram),
            '--disk',
            str(self.flavor.disk),
            '--ephemeral',
            str(self.flavor.ephemeral),
            '--swap',
            str(self.flavor.swap),
            '--vcpus',
            str(self.flavor.vcpus),
            '--rxtx-factor',
            str(self.flavor.rxtx_factor),
            '--private',
            '--description',
            str(self.flavor.description),
            '--project',
            self.project.id,
            '--property',
            'key1=value1',
            '--property',
            'key2=value2',
            self.flavor.name,
        ]
        verifylist = [
            ('ram', self.flavor.ram),
            ('disk', self.flavor.disk),
            ('ephemeral', self.flavor.ephemeral),
            ('swap', self.flavor.swap),
            ('vcpus', self.flavor.vcpus),
            ('rxtx_factor', self.flavor.rxtx_factor),
            ('public', False),
            ('description', self.flavor.description),
            ('project', self.project.id),
            ('properties', {'key1': 'value1', 'key2': 'value2'}),
            ('name', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        args = {
            'name': self.flavor.name,
            'ram': self.flavor.ram,
            'vcpus': self.flavor.vcpus,
            'disk': self.flavor.disk,
            'id': None,
            'ephemeral': self.flavor.ephemeral,
            'swap': self.flavor.swap,
            'rxtx_factor': self.flavor.rxtx_factor,
            'is_public': False,
            'description': self.flavor.description,
        }

        props = {'key1': 'value1', 'key2': 'value2'}

        # SDK updates the flavor object instance. In order to make the
        # verification clear and preciese let's create new flavor and change
        # expected props this way
        create_flavor = _flavor.Flavor(**self.flavor)
        expected_flavor = _flavor.Flavor(**self.flavor)
        expected_flavor.extra_specs = props
        expected_flavor.is_public = False
        # convert expected data tuple to list to be able to modify it
        cmp_data = list(self.data_private)
        cmp_data[7] = format_columns.DictColumn(props)
        self.compute_client.create_flavor.return_value = create_flavor
        self.compute_client.create_flavor_extra_specs.return_value = (
            expected_flavor
        )

        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.create_flavor.assert_called_once_with(**args)
        self.compute_client.flavor_add_tenant_access.assert_called_with(
            self.flavor.id,
            self.project.id,
        )
        self.compute_client.create_flavor_extra_specs.assert_called_with(
            create_flavor, props
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(cmp_data, data)

    def test_public_flavor_create_with_project(self):
        arglist = [
            '--project',
            self.project.id,
            self.flavor.name,
        ]
        verifylist = [
            ('project', self.project.id),
            ('name', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_flavor_create_no_options(self):
        arglist = []
        verifylist = None
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_flavor_create_with_description(self):
        self.set_compute_api_version('2.55')

        arglist = [
            '--id',
            self.flavor.id,
            '--ram',
            str(self.flavor.ram),
            '--disk',
            str(self.flavor.disk),
            '--ephemeral',
            str(self.flavor.ephemeral),
            '--swap',
            str(self.flavor.swap),
            '--vcpus',
            str(self.flavor.vcpus),
            '--rxtx-factor',
            str(self.flavor.rxtx_factor),
            '--public',
            '--description',
            'fake description',
            self.flavor.name,
        ]
        verifylist = [
            ('id', self.flavor.id),
            ('ram', self.flavor.ram),
            ('disk', self.flavor.disk),
            ('ephemeral', self.flavor.ephemeral),
            ('swap', self.flavor.swap),
            ('vcpus', self.flavor.vcpus),
            ('rxtx_factor', self.flavor.rxtx_factor),
            ('public', True),
            ('description', 'fake description'),
            ('name', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        args = {
            'name': self.flavor.name,
            'ram': self.flavor.ram,
            'vcpus': self.flavor.vcpus,
            'disk': self.flavor.disk,
            'id': self.flavor.id,
            'ephemeral': self.flavor.ephemeral,
            'swap': self.flavor.swap,
            'rxtx_factor': self.flavor.rxtx_factor,
            'is_public': True,
            'description': 'fake description',
        }

        self.compute_client.create_flavor.assert_called_once_with(**args)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_flavor_create_with_description_pre_v255(self):
        self.set_compute_api_version('2.54')

        arglist = [
            '--id',
            self.flavor.id,
            '--ram',
            str(self.flavor.ram),
            '--vcpus',
            str(self.flavor.vcpus),
            '--description',
            'description',
            self.flavor.name,
        ]
        verifylist = [
            ('ram', self.flavor.ram),
            ('vcpus', self.flavor.vcpus),
            ('description', 'description'),
            ('name', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_flavor_create_with_rxtx_factor_post_v2102(self):
        self.set_compute_api_version('2.102')

        arglist = [
            '--id',
            self.flavor.id,
            '--ram',
            str(self.flavor.ram),
            '--vcpus',
            str(self.flavor.vcpus),
            '--rxtx-factor',
            '1.0',
            self.flavor.name,
        ]
        verifylist = [
            ('ram', self.flavor.ram),
            ('vcpus', self.flavor.vcpus),
            ('rxtx_factor', 1.0),
            ('name', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestFlavorDelete(TestFlavor):
    def setUp(self):
        super().setUp()

        self.flavors = list(
            sdk_fakes.generate_fake_resources(_flavor.Flavor, 2)
        )

        self.compute_client.delete_flavor.return_value = None

        self.cmd = flavor.DeleteFlavor(self.app, None)

    def test_flavor_delete(self):
        arglist = [self.flavors[0].id]
        verifylist = [
            ('flavor', [self.flavors[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.compute_client.find_flavor.return_value = self.flavors[0]

        result = self.cmd.take_action(parsed_args)

        self.compute_client.find_flavor.assert_called_with(
            self.flavors[0].id, ignore_missing=False
        )
        self.compute_client.delete_flavor.assert_called_with(
            self.flavors[0].id
        )
        self.assertIsNone(result)

    def test_delete_multiple_flavors(self):
        arglist = []
        for f in self.flavors:
            arglist.append(f.id)
        verifylist = [
            ('flavor', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.compute_client.find_flavor.side_effect = self.flavors

        result = self.cmd.take_action(parsed_args)

        find_calls = [
            mock.call(i.id, ignore_missing=False) for i in self.flavors
        ]
        delete_calls = [mock.call(i.id) for i in self.flavors]
        self.compute_client.find_flavor.assert_has_calls(find_calls)
        self.compute_client.delete_flavor.assert_has_calls(delete_calls)
        self.assertIsNone(result)

    def test_multi_flavors_delete_with_exception(self):
        arglist = [
            self.flavors[0].id,
            'unexist_flavor',
        ]
        verifylist = [('flavor', [self.flavors[0].id, 'unexist_flavor'])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.compute_client.find_flavor.side_effect = [
            self.flavors[0],
            sdk_exceptions.ResourceNotFound,
        ]

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 flavors failed to delete.', str(e))

        find_calls = [
            mock.call(self.flavors[0].id, ignore_missing=False),
            mock.call('unexist_flavor', ignore_missing=False),
        ]
        delete_calls = [mock.call(self.flavors[0].id)]
        self.compute_client.find_flavor.assert_has_calls(find_calls)
        self.compute_client.delete_flavor.assert_has_calls(delete_calls)


class TestFlavorList(TestFlavor):
    def setUp(self):
        super().setUp()

        self._flavor = sdk_fakes.generate_fake_resource(
            _flavor.Flavor, extra_specs={'property': 'value'}
        )

        self.columns = (
            'ID',
            'Name',
            'RAM',
            'Disk',
            'Ephemeral',
            'VCPUs',
            'Is Public',
        )
        self.columns_long = (
            *self.columns,
            'Swap',
            'RXTX Factor',
            'Properties',
        )

        self.data = (
            (
                self._flavor.id,
                self._flavor.name,
                self._flavor.ram,
                self._flavor.disk,
                self._flavor.ephemeral,
                self._flavor.vcpus,
                self._flavor.is_public,
            ),
        )
        self.data_long = (
            self.data[0]
            + (
                self._flavor.swap,
                self._flavor.rxtx_factor,
                format_columns.DictColumn(self._flavor.extra_specs),
            ),
        )

        self.compute_client.flavors.side_effect = [[self._flavor], []]

        self.cmd = flavor.ListFlavor(self.app, None)

    def test_flavor_list_no_options(self):
        arglist = []
        verifylist = [
            ('public', True),
            ('all', False),
            ('long', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_public': True,
        }

        self.compute_client.flavors.assert_called_with(**kwargs)
        self.compute_client.fetch_flavor_extra_specs.assert_not_called()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

    def test_flavor_list_all_flavors(self):
        arglist = [
            '--all',
        ]
        verifylist = [
            ('all', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_public': None,
        }

        self.compute_client.flavors.assert_called_with(**kwargs)
        self.compute_client.fetch_flavor_extra_specs.assert_not_called()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

    def test_flavor_list_private_flavors(self):
        arglist = [
            '--private',
        ]
        verifylist = [
            ('public', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_public': False,
        }

        self.compute_client.flavors.assert_called_with(**kwargs)
        self.compute_client.fetch_flavor_extra_specs.assert_not_called()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

    def test_flavor_list_public_flavors(self):
        arglist = [
            '--public',
        ]
        verifylist = [
            ('public', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_public': True,
        }

        self.compute_client.flavors.assert_called_with(**kwargs)
        self.compute_client.fetch_flavor_extra_specs.assert_not_called()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

    def test_flavor_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_public': True,
        }

        self.compute_client.flavors.assert_called_with(**kwargs)
        self.compute_client.fetch_flavor_extra_specs.assert_not_called()

        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, tuple(data))

    def test_flavor_list_long_no_extra_specs(self):
        # use flavor with no extra specs for this test
        flavor = sdk_fakes.generate_fake_resource(
            _flavor.Flavor, extra_specs={}
        )
        self.data = (
            (
                flavor.id,
                flavor.name,
                flavor.ram,
                flavor.disk,
                flavor.ephemeral,
                flavor.vcpus,
                flavor.is_public,
            ),
        )
        self.data_long = (
            self.data[0]
            + (
                flavor.swap,
                flavor.rxtx_factor,
                format_columns.DictColumn(flavor.extra_specs),
            ),
        )

        self.compute_client.flavors.side_effect = [[flavor], []]
        self.compute_client.fetch_flavor_extra_specs.return_value = None

        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_public': True,
        }

        self.compute_client.flavors.assert_called_with(**kwargs)
        self.compute_client.fetch_flavor_extra_specs.assert_called_once_with(
            flavor
        )

        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, tuple(data))

    def test_flavor_list_min_disk_min_ram(self):
        arglist = [
            '--min-disk',
            '10',
            '--min-ram',
            '2048',
        ]
        verifylist = [
            ('min_disk', 10),
            ('min_ram', 2048),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_public': True,
            'min_disk': 10,
            'min_ram': 2048,
        }

        self.compute_client.flavors.assert_called_with(**kwargs)
        self.compute_client.fetch_flavor_extra_specs.assert_not_called()

        self.assertEqual(self.columns, columns)
        self.assertEqual(tuple(self.data), tuple(data))


class TestFlavorSet(TestFlavor):
    def setUp(self):
        super().setUp()

        self.flavor = sdk_fakes.generate_fake_resource(
            _flavor.Flavor, is_public=False, extra_specs={'property': 'value'}
        )
        self.project = sdk_fakes.generate_fake_resource(_project.Project)

        self.compute_client.find_flavor.return_value = self.flavor
        self.projects_mock.get.return_value = self.project
        self.cmd = flavor.SetFlavor(self.app, None)

    def test_flavor_set_property(self):
        arglist = ['--property', 'FOO="B A R"', 'baremetal']
        verifylist = [
            ('properties', {'FOO': '"B A R"'}),
            ('flavor', 'baremetal'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.compute_client.find_flavor.assert_called_with(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False
        )
        self.compute_client.create_flavor_extra_specs.assert_called_with(
            self.flavor.id, {'FOO': '"B A R"'}
        )
        self.assertIsNone(result)

    def test_flavor_set_no_property(self):
        arglist = ['--no-property', 'baremetal']
        verifylist = [('no_property', True), ('flavor', 'baremetal')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.compute_client.find_flavor.assert_called_with(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False
        )
        self.compute_client.delete_flavor_extra_specs_property.assert_called_with(
            self.flavor.id, 'property'
        )
        self.assertIsNone(result)

    def test_flavor_set_project(self):
        arglist = [
            '--project',
            self.project.id,
            self.flavor.id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('flavor', self.flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.find_flavor.assert_called_with(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False
        )
        self.compute_client.flavor_add_tenant_access.assert_called_with(
            self.flavor.id,
            self.project.id,
        )
        self.compute_client.create_flavor_extra_specs.assert_not_called()
        self.assertIsNone(result)

    def test_flavor_set_no_project(self):
        arglist = [
            '--project',
            self.flavor.id,
        ]
        verifylist = [
            ('project', None),
            ('flavor', self.flavor.id),
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_flavor_set_no_flavor(self):
        arglist = [
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('project', self.project.id),
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_flavor_set_with_unexist_flavor(self):
        self.compute_client.find_flavor.side_effect = [
            sdk_exceptions.ResourceNotFound()
        ]

        arglist = [
            '--project',
            self.project.id,
            'unexist_flavor',
        ]
        verifylist = [
            ('project', self.project.id),
            ('flavor', 'unexist_flavor'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_flavor_set_nothing(self):
        arglist = [
            self.flavor.id,
        ]
        verifylist = [
            ('flavor', self.flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.compute_client.find_flavor.assert_called_with(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False
        )
        self.compute_client.flavor_add_tenant_access.assert_not_called()
        self.assertIsNone(result)

    def test_flavor_set_description(self):
        self.set_compute_api_version('2.55')

        arglist = [
            '--description',
            'description',
            self.flavor.id,
        ]
        verifylist = [
            ('description', 'description'),
            ('flavor', self.flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.compute_client.update_flavor.assert_called_with(
            flavor=self.flavor.id, description='description'
        )
        self.assertIsNone(result)

    def test_flavor_set_description_pre_v254(self):
        self.set_compute_api_version('2.54')

        arglist = [
            '--description',
            'description',
            self.flavor.id,
        ]
        verifylist = [
            ('description', 'description'),
            ('flavor', self.flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_flavor_set_description_using_name(self):
        self.set_compute_api_version('2.55')

        arglist = [
            '--description',
            'description',
            self.flavor.name,
        ]
        verifylist = [
            ('description', 'description'),
            ('flavor', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.compute_client.update_flavor.assert_called_with(
            flavor=self.flavor.id, description='description'
        )
        self.assertIsNone(result)

    def test_flavor_set_description_using_name_pre_v255(self):
        self.set_compute_api_version('2.54')

        arglist = [
            '--description',
            'description',
            self.flavor.name,
        ]
        verifylist = [
            ('description', 'description'),
            ('flavor', self.flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestFlavorShow(TestFlavor):
    def setUp(self):
        super().setUp()

        self.flavor_access = compute_fakes.create_one_flavor_access()
        self.flavor = sdk_fakes.generate_fake_resource(_flavor.Flavor)

        self.columns = (
            'OS-FLV-DISABLED:disabled',
            'OS-FLV-EXT-DATA:ephemeral',
            'access_project_ids',
            'description',
            'disk',
            'id',
            'name',
            'os-flavor-access:is_public',
            'properties',
            'ram',
            'rxtx_factor',
            'swap',
            'vcpus',
        )
        self.data = (
            self.flavor.is_disabled,
            self.flavor.ephemeral,
            None,
            self.flavor.description,
            self.flavor.disk,
            self.flavor.id,
            self.flavor.name,
            self.flavor.is_public,
            format_columns.DictColumn(self.flavor.extra_specs),
            self.flavor.ram,
            self.flavor.rxtx_factor,
            self.flavor.swap,
            self.flavor.vcpus,
        )
        self.compute_client.find_flavor.return_value = self.flavor
        self.compute_client.get_flavor_access.return_value = [
            self.flavor_access
        ]
        self.cmd = flavor.ShowFlavor(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should boil here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_public_flavor_show(self):
        arglist = [
            self.flavor.name,
        ]
        verifylist = [
            ('flavor', self.flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_private_flavor_show(self):
        private_flavor = sdk_fakes.generate_fake_resource(
            _flavor.Flavor, is_public=False
        )
        self.compute_client.find_flavor.return_value = private_flavor

        arglist = [
            private_flavor.name,
        ]
        verifylist = [
            ('flavor', private_flavor.name),
        ]

        data_with_project = (
            private_flavor.is_disabled,
            private_flavor.ephemeral,
            [self.flavor_access['tenant_id']],
            private_flavor.description,
            private_flavor.disk,
            private_flavor.id,
            private_flavor.name,
            private_flavor.is_public,
            format_columns.DictColumn(private_flavor.extra_specs),
            private_flavor.ram,
            private_flavor.rxtx_factor,
            private_flavor.swap,
            private_flavor.vcpus,
        )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.get_flavor_access.assert_called_with(
            flavor=private_flavor.id
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(data_with_project, data)


class TestFlavorUnset(TestFlavor):
    def setUp(self):
        super().setUp()

        self.flavor = sdk_fakes.generate_fake_resource(
            _flavor.Flavor, is_public=False
        )
        self.project = sdk_fakes.generate_fake_resource(_project.Project)

        self.compute_client.find_flavor.return_value = self.flavor
        self.projects_mock.get.return_value = self.project

        self.cmd = flavor.UnsetFlavor(self.app, None)

    def test_flavor_unset_property(self):
        arglist = ['--property', 'property', 'baremetal']
        verifylist = [
            ('properties', ['property']),
            ('flavor', 'baremetal'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.compute_client.find_flavor.assert_called_with(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False
        )
        self.compute_client.delete_flavor_extra_specs_property.assert_called_with(
            self.flavor.id, 'property'
        )
        self.compute_client.flavor_remove_tenant_access.assert_not_called()
        self.assertIsNone(result)

    def test_flavor_unset_properties(self):
        arglist = [
            '--property',
            'property1',
            '--property',
            'property2',
            'baremetal',
        ]
        verifylist = [
            ('properties', ['property1', 'property2']),
            ('flavor', 'baremetal'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.compute_client.find_flavor.assert_called_with(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False
        )
        self.compute_client.delete_flavor_extra_specs_property.assert_has_calls(
            [
                mock.call(self.flavor.id, 'property1'),
                mock.call(self.flavor.id, 'property2'),
            ]
        )
        self.compute_client.flavor_remove_tenant_access.assert_not_called()

    def test_flavor_unset_project(self):
        arglist = [
            '--project',
            self.project.id,
            self.flavor.id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('flavor', self.flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.compute_client.find_flavor.assert_called_with(
            parsed_args.flavor, get_extra_specs=True, ignore_missing=False
        )
        self.compute_client.flavor_remove_tenant_access.assert_called_with(
            self.flavor.id,
            self.project.id,
        )
        self.compute_client.delete_flavor_extra_specs_property.assert_not_called()
        self.assertIsNone(result)

    def test_flavor_unset_no_project(self):
        arglist = [
            '--project',
            self.flavor.id,
        ]
        verifylist = [
            ('project', None),
            ('flavor', self.flavor.id),
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_flavor_unset_no_flavor(self):
        arglist = [
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('project', self.project.id),
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_flavor_unset_with_unexist_flavor(self):
        self.compute_client.find_flavor.side_effect = [
            sdk_exceptions.ResourceNotFound
        ]

        arglist = [
            '--project',
            self.project.id,
            'unexist_flavor',
        ]
        verifylist = [
            ('project', self.project.id),
            ('flavor', 'unexist_flavor'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_flavor_unset_nothing(self):
        arglist = [
            self.flavor.id,
        ]
        verifylist = [
            ('flavor', self.flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.compute_client.flavor_remove_tenant_access.assert_not_called()



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""IP Floating action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import exceptions as sdk_exceptions
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import tags as _tag

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common

LOG = logging.getLogger(__name__)

_formatters = {
    'port_details': format_columns.DictColumn,
}


def _get_network_columns(item: Any) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


def _get_columns(item: Any) -> tuple[str, ...]:
    columns = list(item.keys())
    return tuple(sorted(columns))


def _get_attrs(
    client_manager: Any, parsed_args: argparse.Namespace
) -> dict[str, Any]:
    attrs = {}
    network_client = client_manager.network

    # Name of a network could be empty string.
    if parsed_args.network is not None:
        network = network_client.find_network(
            parsed_args.network, ignore_missing=False
        )
        attrs['floating_network_id'] = network.id

    if parsed_args.subnet:
        subnet = network_client.find_subnet(
            parsed_args.subnet, ignore_missing=False
        )
        attrs['subnet_id'] = subnet.id

    if parsed_args.port:
        port = network_client.find_port(parsed_args.port, ignore_missing=False)
        attrs['port_id'] = port.id

    if parsed_args.floating_ip_address:
        attrs['floating_ip_address'] = parsed_args.floating_ip_address

    if parsed_args.fixed_ip_address:
        attrs['fixed_ip_address'] = parsed_args.fixed_ip_address

    if parsed_args.qos_policy:
        attrs['qos_policy_id'] = network_client.find_qos_policy(
            parsed_args.qos_policy, ignore_missing=False
        ).id

    if parsed_args.description is not None:
        attrs['description'] = parsed_args.description

    if parsed_args.project:
        identity_client = client_manager.identity
        project_id = identity_common.find_project(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        ).id
        attrs['project_id'] = project_id

    if parsed_args.dns_domain:
        attrs['dns_domain'] = parsed_args.dns_domain

    if parsed_args.dns_name:
        attrs['dns_name'] = parsed_args.dns_name

    return attrs


class CreateFloatingIP(command.ShowOne, common.NeutronCommandWithExtraArgs):
    _description = _("Create floating IP")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar='<network>',
            help=_("Network to allocate floating IP from (name or ID)"),
        )
        parser.add_argument(
            '--subnet',
            metavar='<subnet>',
            help=_(
                "Subnet on which you want to create the floating IP "
                "(name or ID)"
            ),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help=_("Port to be associated with the floating IP (name or ID)"),
        )
        parser.add_argument(
            '--floating-ip-address',
            metavar='<ip-address>',
            dest='floating_ip_address',
            help=_("Floating IP address"),
        )
        parser.add_argument(
            '--fixed-ip-address',
            metavar='<ip-address>',
            dest='fixed_ip_address',
            help=_("Fixed IP address mapped to the floating IP"),
        )
        parser.add_argument(
            '--qos-policy',
            metavar='<qos-policy>',
            help=_("Attach QoS policy to the floating IP (name or ID)"),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Set floating IP description'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("Owner's project (name or ID)"),
        )
        parser.add_argument(
            '--dns-domain',
            metavar='<dns-domain>',
            dest='dns_domain',
            help=_("Set DNS domain for this floating IP"),
        )
        parser.add_argument(
            '--dns-name',
            metavar='<dns-name>',
            dest='dns_name',
            help=_("Set DNS name for this floating IP"),
        )

        identity_common.add_project_domain_option_to_parser(parser)
        _tag.add_tag_option_to_parser_for_create(parser, _('floating IP'))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        with common.check_missing_extension_if_error(client, attrs):
            obj = client.create_ip(**attrs)

        # tags cannot be set when created, so tags need to be set later.
        _tag.update_tags_for_set(client, obj, parsed_args)

        display_columns, columns = _get_network_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)


class DeleteFloatingIP(command.Command):
    _description = _("Delete floating IP(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar="<floating-ip>",
            nargs="+",
            help=_("Floating IP(s) to delete (IP address or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for fip in parsed_args.floating_ip:
            try:
                obj = client.find_ip(fip, ignore_missing=False)
                client.delete_ip(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete floating IP with "
                        "name or ID '%(fip)s': %(e)s"
                    ),
                    {'fip': fip, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.floating_ip)
            msg = _(
                "%(result)s of %(total)s floating IPs failed to delete."
            ) % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListFloatingIP(command.Lister):
    # TODO(songminglong): Use SDK resource mapped attribute names once
    # the OSC minimum requirements include SDK 1.0

    _description = _("List floating IP(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--network',
            metavar='<network>',
            dest='networks',
            action='append',
            help=_(
                "List only floating IP(s) with the specified network "
                "(name or ID) "
                "(repeat option to fiter on multiple networks)"
            ),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            dest='ports',
            action='append',
            help=_(
                "List only floating IP(s) with the specified port "
                "(name or ID) "
                "(repeat option to fiter on multiple ports)"
            ),
        )
        parser.add_argument(
            '--fixed-ip-address',
            metavar='<ip-address>',
            help=_(
                "List only floating IP(s) with the specified fixed IP address"
            ),
        )
        parser.add_argument(
            '--floating-ip-address',
            metavar='<ip-address>',
            help=_(
                "List only floating IP(s) with the specified floating IP "
                "address"
            ),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            choices=['ACTIVE', 'DOWN'],
            help=_(
                "List only floating IP(s) with the specified status "
                "('ACTIVE', 'DOWN')"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "List only floating IP(s) with the specified project "
                "(name or ID)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--router',
            metavar='<router>',
            dest='routers',
            action='append',
            help=_(
                "List only floating IP(s) with the specified router "
                "(name or ID) "
                "(repeat option to fiter on multiple routers)"
            ),
        )
        _tag.add_tag_filtering_option_to_parser(parser, _('floating IP'))
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        network_client = self.app.client_manager.network
        identity_client = self.app.client_manager.identity

        columns: tuple[str, ...] = (
            'id',
            'floating_ip_address',
            'fixed_ip_address',
            'port_id',
            'floating_network_id',
            'project_id',
        )
        headers: tuple[str, ...] = (
            'ID',
            'Floating IP Address',
            'Fixed IP Address',
            'Port',
            'Floating Network',
            'Project',
        )
        if parsed_args.long:
            columns += (
                'router_id',
                'status',
                'description',
                'tags',
                'dns_name',
                'dns_domain',
            )
            headers += (
                'Router',
                'Status',
                'Description',
                'Tags',
                'DNS Name',
                'DNS Domain',
            )

        query = {}

        if parsed_args.networks is not None:
            network_ids = []
            for network in parsed_args.networks:
                network_id = network_client.find_network(
                    network, ignore_missing=False
                ).id
                network_ids.append(network_id)
            query['floating_network_id'] = network_ids

        if parsed_args.ports is not None:
            port_ids = []
            for port in parsed_args.ports:
                port_id = network_client.find_port(
                    port, ignore_missing=False
                ).id
                port_ids.append(port_id)
            query['port_id'] = port_ids

        if parsed_args.fixed_ip_address is not None:
            query['fixed_ip_address'] = parsed_args.fixed_ip_address

        if parsed_args.floating_ip_address is not None:
            query['floating_ip_address'] = parsed_args.floating_ip_address

        if parsed_args.status:
            query['status'] = parsed_args.status

        if parsed_args.project is not None:
            project = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            query['project_id'] = project.id

        if parsed_args.routers is not None:
            router_ids = []
            for router in parsed_args.routers:
                router_id = network_client.find_router(
                    router, ignore_missing=False
                ).id
                router_ids.append(router_id)
            query['router_id'] = router_ids

        _tag.get_tag_filtering_args(parsed_args, query)

        try:
            data = list(network_client.ips(**query))
        except sdk_exceptions.NotFoundException:
            data = []

        return (
            headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetFloatingIP(common.NeutronCommandWithExtraArgs):
    _description = _("Set floating IP Properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating-ip>',
            help=_("Floating IP to modify (IP address or ID)"),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help=_("Associate the floating IP with port (name or ID)"),
        )
        parser.add_argument(
            '--fixed-ip-address',
            metavar='<ip-address>',
            dest='fixed_ip_address',
            help=_(
                "Fixed IP of the port (required only if port has multiple IPs)"
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Set floating IP description'),
        )
        qos_policy_group = parser.add_mutually_exclusive_group()
        qos_policy_group.add_argument(
            '--qos-policy',
            metavar='<qos-policy>',
            help=_("Attach QoS policy to the floating IP (name or ID)"),
        )
        qos_policy_group.add_argument(
            '--no-qos-policy',
            action='store_true',
            help=_("Remove the QoS policy attached to the floating IP"),
        )

        _tag.add_tag_option_to_parser_for_set(parser, _('floating IP'))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )

        attrs: dict[str, Any] = {}

        if parsed_args.port:
            port = client.find_port(parsed_args.port, ignore_missing=False)
            attrs['port_id'] = port.id

        if parsed_args.fixed_ip_address:
            attrs['fixed_ip_address'] = parsed_args.fixed_ip_address

        if parsed_args.description:
            attrs['description'] = parsed_args.description

        if parsed_args.qos_policy:
            attrs['qos_policy_id'] = client.find_qos_policy(
                parsed_args.qos_policy, ignore_missing=False
            ).id

        if 'no_qos_policy' in parsed_args and parsed_args.no_qos_policy:
            attrs['qos_policy_id'] = None

        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        if attrs:
            client.update_ip(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_set(client, obj, parsed_args)


class ShowFloatingIP(command.ShowOne):
    _description = _("Display floating IP details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar="<floating-ip>",
            help=_("Floating IP to display (IP address or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )
        display_columns, columns = _get_network_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)


class UnsetFloatingIP(common.NeutronCommandWithExtraArgs):
    _description = _("Unset floating IP Properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating-ip>',
            help=_("Floating IP to disassociate (IP address or ID)"),
        )
        parser.add_argument(
            '--port',
            action='store_true',
            default=False,
            help=_("Disassociate any port associated with the floating IP"),
        )
        parser.add_argument(
            '--qos-policy',
            action='store_true',
            default=False,
            help=_("Remove the QoS policy attached to the floating IP"),
        )
        _tag.add_tag_option_to_parser_for_unset(parser, _('floating IP'))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )
        attrs: dict[str, None] = {}
        if parsed_args.port:
            attrs['port_id'] = None
        if parsed_args.qos_policy:
            attrs['qos_policy_id'] = None
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        if attrs:
            client.update_ip(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_unset(client, obj, parsed_args)



# Copyright 2016 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy
import re
from unittest import mock

from openstack.network.v2 import firewall_group
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network.v2.fwaas import group as fwaas_group
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as test_utils


CONVERT_MAP = {
    'ingress_firewall_policy': 'ingress_firewall_policy_id',
    'egress_firewall_policy': 'egress_firewall_policy_id',
    'no_ingress_firewall_policy': 'ingress_firewall_policy_id',
    'no_egress_firewall_policy': 'egress_firewall_policy_id',
    'project': 'project_id',
    'port': 'ports',
}


def _generate_response(source=None, data=None):
    source = source if source else {}
    up = {
        'admin_state_up': fwaas_group.AdminStateColumn(
            source['admin_state_up']
        )
    }
    if data:
        up.append(data)
    source.update(up)
    return source


def _generate_req_and_res(verifylist, response):
    request = dict(verifylist)
    for key, val in verifylist:
        del request[key]
        if re.match('^no_', key) and val is True:
            new_value = None
        elif val is True or val is False:
            new_value = val
        elif key in ('name', 'description'):
            new_value = val
        else:
            new_value = val
        converted = CONVERT_MAP.get(key, key)
        request[converted] = new_value
        response[converted] = new_value
    return request, response


class TestFirewallGroup(network_fakes.TestNetworkV2):
    def check_results(self, headers, data, exp_req, is_list=False):
        if is_list:
            req_body = {self.res_plural: list(exp_req)}
        else:
            req_body = exp_req
        self.mocked.assert_called_once_with(**req_body)
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))

    def setUp(self):
        super().setUp()

        self.resource = sdk_fakes.generate_fake_resource(
            firewall_group.FirewallGroup
        )

        def _find_resource(*args, **kwargs):
            return self.resource

        self.network_client.find_firewall_group.side_effect = _find_resource
        self.identity_client.projects.get.side_effect = lambda x: mock.Mock(
            id=x
        )
        self.res = 'firewall_group'
        self.res_plural = 'firewall_groups'
        self.list_headers = (
            'ID',
            'Name',
            'Ingress Policy ID',
            'Egress Policy ID',
        )
        self.list_data = (
            self.resource['id'],
            self.resource['name'],
            self.resource['ingress_firewall_policy_id'],
            self.resource['egress_firewall_policy_id'],
        )
        self.headers = tuple(
            (
                *self.list_headers,
                'Description',
                'Status',
                'Ports',
                'State',
                'Shared',
                'Project',
            )
        )
        self.data = _generate_response(self.resource)
        self.ordered_headers = copy.deepcopy(tuple(sorted(self.headers)))
        self.expected_data = (
            self.resource['description'],
            self.resource['egress_firewall_policy_id'],
            self.resource['id'],
            self.resource['ingress_firewall_policy_id'],
            self.resource['name'],
            self.resource['ports'],
            self.resource['project_id'],
            self.resource['shared'],
            fwaas_group.AdminStateColumn(self.resource['admin_state_up']),
            self.resource['status'],
        )
        self.ordered_columns = (
            'description',
            'egress_firewall_policy_id',
            'id',
            'ingress_firewall_policy_id',
            'name',
            'ports',
            'project_id',
            'shared',
            'admin_state_up',
            'status',
        )


class TestCreateFirewallGroup(TestFirewallGroup):
    def setUp(self):
        super().setUp()
        self.network_client.create_firewall_group.return_value = self.resource
        self.mocked = self.network_client.create_firewall_group
        self.cmd = fwaas_group.CreateFirewallGroup(self.app, None)

    def _update_expect_response(self, request, response):
        """Set expected request and response

        :param request
            A dictionary of request body(dict of verifylist)
        :param response
            A OrderedDict of request body
        """
        # Update response body
        self.network_client.create_firewall_group.return_value = response
        # Update response(finally returns 'data')
        self.data = _generate_response(source=response)
        self.expected_data = response

    def test_create_with_no_option(self):
        # firewall_group-create with mandatory (none) params.
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))

    def test_create_with_port(self):
        # firewall_group-create with 'port'
        port_id = 'id_for_port'

        def _mock_find(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_port.side_effect = _mock_find
        arglist = ['--port', port_id]
        verifylist = [('port', [port_id])]
        request, response = _generate_req_and_res(verifylist, self.resource)
        self._update_expect_response(request, response)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.check_results(headers, data, request)

    def test_create_with_ingress_policy(self):
        ingress_policy = 'my-ingress-policy'

        def _mock_port_fwg(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_policy.side_effect = _mock_port_fwg

        arglist = ['--ingress-firewall-policy', ingress_policy]
        verifylist = [('ingress_firewall_policy', ingress_policy)]
        request, response = _generate_req_and_res(verifylist, self.resource)
        self._update_expect_response(request, response)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)
        self.network_client.find_firewall_policy.assert_called_once_with(
            ingress_policy, ignore_missing=False
        )

        self.check_results(headers, data, request)

    def test_create_with_egress_policy(self):
        egress_policy = 'my-egress-policy'

        def _mock_find(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_group.side_effect = _mock_find
        self.network_client.find_firewall_policy.side_effect = _mock_find

        arglist = ['--egress-firewall-policy', egress_policy]
        verifylist = [('egress_firewall_policy', egress_policy)]
        request, response = _generate_req_and_res(verifylist, self.resource)
        self._update_expect_response(request, response)
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.find_firewall_policy.assert_called_once_with(
            egress_policy, ignore_missing=False
        )
        self.check_results(headers, data, request)

    def test_create_with_all_params(self):
        name = 'my-name'
        description = 'my-desc'
        ingress_policy = 'my-ingress-policy'
        egress_policy = 'my-egress-policy'

        def _mock_find(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_policy.side_effect = _mock_find
        port = 'port'
        self.network_client.find_port.side_effect = _mock_find
        project_id = 'my-project'
        arglist = [
            '--name',
            name,
            '--description',
            description,
            '--ingress-firewall-policy',
            ingress_policy,
            '--egress-firewall-policy',
            egress_policy,
            '--port',
            port,
            '--project',
            project_id,
            '--share',
            '--disable',
        ]
        verifylist = [
            ('name', name),
            ('description', description),
            ('ingress_firewall_policy', ingress_policy),
            ('egress_firewall_policy', egress_policy),
            ('port', [port]),
            ('shared', True),
            ('project', project_id),
            ('admin_state_up', False),
        ]
        request, response = _generate_req_and_res(verifylist, self.resource)
        self._update_expect_response(request, response)
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.check_results(headers, data, request)

    def test_create_with_shared_and_no_share(self):
        arglist = [
            '--share',
            '--no-share',
        ]
        verifylist = [
            ('shared', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_ports_and_no(self):
        port = 'my-port'
        arglist = [
            '--port',
            port,
            '--no-port',
        ]
        verifylist = [
            ('port', [port]),
            ('no_port', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_ingress_policy_and_no(self):
        policy = 'my-policy'
        arglist = [
            '--ingress-firewall-policy',
            policy,
            '--no-ingress-firewall-policy',
        ]
        verifylist = [
            ('ingress_firewall_policy', policy),
            ('no_ingress_firewall_policy', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_egress_policy_and_no(self):
        policy = 'my-policy'
        arglist = [
            '--egress-firewall-policy',
            policy,
            '--no-egress-firewall-policy',
        ]
        verifylist = [
            ('egress_firewall_policy', policy),
            ('no_egress_firewall_policy', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestListFirewallGroup(TestFirewallGroup):
    def setUp(self):
        super().setUp()
        self.network_client.firewall_groups.return_value = [self.resource]
        self.mocked = self.network_client.firewall_groups
        self.cmd = fwaas_group.ListFirewallGroup(self.app, None)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.list_headers), headers)
        self.assertEqual([self.list_data], list(data))

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)


class TestShowFirewallGroup(TestFirewallGroup):
    def setUp(self):
        super().setUp()
        self.network_client.get_firewall_group.return_value = self.resource
        self.mocked = self.network_client.get_firewall_group
        self.cmd = fwaas_group.ShowFirewallGroup(self.app, None)

    def test_show_filtered_by_id_or_name(self):
        target = self.resource['id']

        def _mock_fwaas(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_group.side_effect = _mock_fwaas

        arglist = [target]
        verifylist = [(self.res, target)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target)
        self.assertEqual(self.ordered_headers, headers)


class TestSetFirewallGroup(TestFirewallGroup):
    def setUp(self):
        super().setUp()
        self.resource['ports'] = ['old_port']
        self.network_client.update_firewall_group.return_value = {
            self.res: self.resource
        }
        self.mocked = self.network_client.update_firewall_group

        def _mock_find_port(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_port.side_effect = _mock_find_port

        self.cmd = fwaas_group.SetFirewallGroup(self.app, None)

    def _update_expect_response(self, request, response):
        """Set expected request and response

        :param request
            A dictionary of request body(dict of verifylist)
        :param response
            A OrderedDict of request body
        """
        # Update response(finally returns 'data')
        self.data = _generate_response(source=response)
        self.ordered_data = tuple(
            response[column] for column in self.ordered_columns
        )

    def test_set_name(self):
        target = self.resource['id']
        update = 'change'
        arglist = [target, '--name', update]
        verifylist = [
            (self.res, target),
            ('name', update),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **{'name': update})
        self.assertIsNone(result)

    def test_set_options(self):
        target = self.resource['id']
        updated_desc = 'change-desc'
        arglist = [target, '--description', updated_desc, '--share']
        verifylist = [
            (self.res, target),
            ('description', updated_desc),
            ('shared', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, shared=True, description=updated_desc
        )
        self.assertIsNone(result)

    def test_set_ingress_policy_and_egress_policy(self):
        target = self.resource['id']
        ingress_policy = 'ingress_policy'
        egress_policy = 'egress_policy'

        def _mock_fwg_policy(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_group.side_effect = _mock_fwg_policy
        self.network_client.find_firewall_policy.side_effect = _mock_fwg_policy

        arglist = [
            target,
            '--ingress-firewall-policy',
            ingress_policy,
            '--egress-firewall-policy',
            egress_policy,
        ]
        verifylist = [
            (self.res, target),
            ('ingress_firewall_policy', ingress_policy),
            ('egress_firewall_policy', egress_policy),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target,
            **{
                'ingress_firewall_policy_id': ingress_policy,
                'egress_firewall_policy_id': egress_policy,
            },
        )
        self.assertIsNone(result)

    def test_set_port(self):
        target = self.resource['id']
        port1 = 'additional_port1'
        port2 = 'additional_port2'

        def _mock_port_fwg(*args, **kwargs):
            return mock.Mock(id=args[0], ports=self.resource['ports'])

        self.network_client.find_firewall_group.side_effect = _mock_port_fwg
        self.network_client.find_port.side_effect = _mock_port_fwg

        arglist = [
            target,
            '--port',
            port1,
            '--port',
            port2,
        ]
        verifylist = [
            (self.res, target),
            ('port', [port1, port2]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        expect = {'ports': sorted(self.resource['ports'] + [port1, port2])}
        self.mocked.assert_called_once_with(target, **expect)
        self.assertEqual(2, self.network_client.find_firewall_group.call_count)
        self.assertIsNone(result)

    def test_set_no_port(self):
        # firewall_group-update myid --policy newpolicy.
        target = self.resource['id']
        arglist = [target, '--no-port']
        verifylist = [
            (self.res, target),
            ('no_port', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **{'ports': []})
        self.assertIsNone(result)

    def test_set_admin_state(self):
        target = self.resource['id']
        arglist = [target, '--enable']
        verifylist = [
            (self.res, target),
            ('admin_state_up', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **{'admin_state_up': True})
        self.assertIsNone(result)

    def test_set_shared(self):
        target = self.resource['id']
        arglist = [target, '--share']
        verifylist = [
            (self.res, target),
            ('shared', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **{'shared': True})
        self.assertIsNone(result)

    def test_set_no_share(self):
        target = self.resource['id']
        arglist = [target, '--no-share']
        verifylist = [
            (self.res, target),
            ('shared', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **{'shared': False})
        self.assertIsNone(result)

    def test_set_egress_policy(self):
        target = self.resource['id']
        policy = 'egress_policy'

        def _mock_find_policy(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_policy.side_effect = (
            _mock_find_policy
        )

        arglist = [target, '--egress-firewall-policy', policy]
        verifylist = [
            (self.res, target),
            ('egress_firewall_policy', policy),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, **{'egress_firewall_policy_id': policy}
        )
        self.assertIsNone(result)

    def test_set_no_ingress_policies(self):
        target = self.resource['id']
        arglist = [target, '--no-ingress-firewall-policy']
        verifylist = [
            (self.res, target),
            ('no_ingress_firewall_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, **{'ingress_firewall_policy_id': None}
        )
        self.assertIsNone(result)

    def test_set_no_egress_policies(self):
        target = self.resource['id']
        arglist = [target, '--no-egress-firewall-policy']
        verifylist = [
            (self.res, target),
            ('no_egress_firewall_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, **{'egress_firewall_policy_id': None}
        )
        self.assertIsNone(result)

    def test_set_port_and_no_port(self):
        target = self.resource['id']
        port = 'my-port'
        arglist = [
            target,
            '--port',
            port,
            '--no-port',
        ]
        verifylist = [
            (self.res, target),
            ('port', [port]),
            ('no_port', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.mocked.assert_called_once_with(target, **{'ports': [port]})
        self.assertIsNone(result)

    def test_set_ingress_policy_and_no_ingress_policy(self):
        target = self.resource['id']
        arglist = [
            target,
            '--ingress-firewall-policy',
            'my-ingress',
            '--no-ingress-firewall-policy',
        ]
        verifylist = [
            (self.res, target),
            ('ingress_firewall_policy', 'my-ingress'),
            ('no_ingress_firewall_policy', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_set_egress_policy_and_no_egress_policy(self):
        target = self.resource['id']
        arglist = [
            target,
            '--egress-firewall-policy',
            'my-egress',
            '--no-egress-firewall-policy',
        ]
        verifylist = [
            (self.res, target),
            ('egress_firewall_policy', 'my-egress'),
            ('no_egress_firewall_policy', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_set_and_raises(self):
        self.network_client.update_firewall_group.side_effect = Exception
        target = self.resource['id']
        arglist = [target, '--name', 'my-name']
        verifylist = [(self.res, target), ('name', 'my-name')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestDeleteFirewallGroup(TestFirewallGroup):
    def setUp(self):
        super().setUp()
        # Mock objects
        self.mocked = self.network_client.delete_firewall_group
        self.cmd = fwaas_group.DeleteFirewallGroup(self.app, None)

    def test_delete_with_one_resource(self):
        target = self.resource['id']

        def _mock_fwaas(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_group.side_effect = _mock_fwaas

        arglist = [target]
        verifylist = [(self.res, [target])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target)
        self.assertIsNone(result)

    def test_delete_with_multiple_resources(self):

        def _mock_fwaas(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_group.side_effect = _mock_fwaas

        target1 = 'target1'
        target2 = 'target2'
        arglist = [target1, target2]
        verifylist = [(self.res, [target1, target2])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.assertEqual(2, self.mocked.call_count)
        for idx, reference in enumerate([target1, target2]):
            actual = ''.join(self.mocked.call_args_list[idx][0][0])
            self.assertEqual(reference, actual)

    def test_delete_multiple_with_exception(self):
        target1 = 'target1'
        target2 = 'target2'
        arglist = [target1, target2]
        verifylist = [(self.res, [target1, target2])]

        def _mock_find(*args, **kwargs):
            if args[0] == target2:
                raise Exception('Not found')
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_group.side_effect = _mock_find

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestUnsetFirewallGroup(TestFirewallGroup):
    def setUp(self):
        super().setUp()
        self.resource['ports'] = ['old_port']
        # Mock objects
        self.mocked = self.network_client.update_firewall_group
        self.cmd = fwaas_group.UnsetFirewallGroup(self.app, None)

    def test_unset_shared(self):
        target = self.resource['id']
        arglist = [
            target,
            '--share',
        ]
        verifylist = [
            (self.res, target),
            ('share', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        with mock.patch.object(fwaas_group.LOG, 'warning') as mock_warning:
            result = self.cmd.take_action(parsed_args)
            mock_warning.assert_called_once_with(
                'The --share option is deprecated, please use '
                '"firewall group set --no-share" instead.'
            )
        self.mocked.assert_called_once_with(target, **{'shared': False})
        self.assertIsNone(result)

    def test_unset_ingress_policy(self):
        target = self.resource['id']
        arglist = [
            target,
            '--ingress-firewall-policy',
        ]
        verifylist = [
            (self.res, target),
            ('ingress_firewall_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.mocked.assert_called_once_with(
            target, **{'ingress_firewall_policy_id': None}
        )
        self.assertIsNone(result)

    def test_unset_egress_policy(self):
        target = self.resource['id']
        arglist = [
            target,
            '--egress-firewall-policy',
        ]
        verifylist = [
            (self.res, target),
            ('egress_firewall_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.mocked.assert_called_once_with(
            target, **{'egress_firewall_policy_id': None}
        )
        self.assertIsNone(result)

    def test_unset_enable(self):
        target = self.resource['id']
        arglist = [
            target,
            '--enable',
        ]
        verifylist = [
            (self.res, target),
            ('enable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        with mock.patch.object(fwaas_group.LOG, 'warning') as mock_warning:
            result = self.cmd.take_action(parsed_args)
            mock_warning.assert_called_once_with(
                'The --enable option is deprecated, please use '
                '"firewall group set --disable" instead.'
            )
        self.mocked.assert_called_once_with(
            target, **{'admin_state_up': False}
        )
        self.assertIsNone(result)

    def test_unset_port(self):
        target = self.resource['id']
        port = 'old_port'

        def _mock_port_fwg(*args, **kwargs):
            return mock.Mock(id=args[0], ports=self.resource['ports'])

        self.network_client.find_firewall_group.side_effect = _mock_port_fwg
        self.network_client.find_port.side_effect = _mock_port_fwg

        arglist = [
            target,
            '--port',
            port,
        ]
        verifylist = [
            (self.res, target),
            ('port', [port]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.mocked.assert_called_once_with(target, **{'ports': []})
        self.assertIsNone(result)

    def test_unset_all_port(self):
        target = self.resource['id']
        arglist = [
            target,
            '--all-port',
        ]
        verifylist = [
            (self.res, target),
            ('all_port', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.mocked.assert_called_once_with(target, **{'ports': []})
        self.assertIsNone(result)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstackclient.tests.functional import base


class MetadefObjectTests(base.TestCase):
    def setUp(self):
        super().setUp()
        self.obj_name = self.getUniqueString('metadef-obj')
        self.ns_name = self.getUniqueString('metadef-ns')
        self.openstack(f"image metadef namespace create {self.ns_name}")
        self.addCleanup(
            lambda: self.openstack(
                f"image metadef namespace delete {self.ns_name}"
            )
        )

    def test_metadef_objects(self):
        # CREATE
        created = self.openstack(
            (
                "image metadef object create "
                f"--namespace {self.ns_name} "
                f"{self.obj_name}"
            ),
            parse_output=True,
        )
        self.addCleanup(
            lambda: self.openstack(
                f"image metadef object delete {self.ns_name} {self.obj_name}"
            )
        )
        self.assertEqual(self.obj_name, created["name"])
        self.assertEqual(self.ns_name, created["namespace_name"])

        # UPDATE
        new_name = f"{self.obj_name}-updated"
        self.openstack(
            "image metadef object update "
            f"{self.ns_name} {self.obj_name} "
            f"--name {new_name}"
        )
        self.obj_name = new_name

        # READ (get)
        shown = self.openstack(
            f"image metadef object show {self.ns_name} {self.obj_name}",
            parse_output=True,
        )
        self.assertEqual(self.obj_name, shown["name"])
        self.assertEqual(self.ns_name, shown["namespace_name"])

        # READ (list)
        rows = self.openstack(
            f"image metadef object list {self.ns_name}",
            parse_output=True,
        )
        names = {row["name"] for row in rows}
        self.assertIn(self.obj_name, names)



#   Copyright 2014 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 federation mapping action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import json
import logging
from typing import Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class _RulesReader:
    _description = _("Helper class capable of reading rules from files")

    def _read_rules(self, path: str) -> Any:
        """Read and parse rules from path

        Expect the file to contain a valid JSON structure.

        :param path: path to the file
        :return: loaded and valid dictionary with rules
        :raises exception.CommandError: In case the file cannot be
            accessed or the content is not a valid JSON.

        Example of the content of the file:
            [
                {
                    "local": [
                        {
                            "group": {
                                "id": "85a868"
                            }
                        }
                    ],
                    "remote": [
                        {
                            "type": "orgPersonType",
                            "any_one_of": [
                                "Employee"
                            ]
                        },
                        {
                            "type": "sn",
                            "any_one_of": [
                                "Young"
                            ]
                        }
                    ]
                }
            ]

        """
        blob = utils.read_blob_file_contents(path)
        try:
            rules = json.loads(blob)
        except ValueError as e:
            msg = _(
                "An error occurred when reading rules from file "
                "%(path)s: %(error)s"
            ) % {"path": path, "error": e}
            raise exceptions.CommandError(msg)
        else:
            return rules

    @staticmethod
    def add_federated_schema_version_option(
        parser: argparse.ArgumentParser,
    ) -> None:
        parser.add_argument(
            '--schema-version',
            metavar='<schema_version>',
            required=False,
            default=None,
            help=_(
                "The federated attribute mapping schema version. The "
                "default value on the client side is 'None'; however, that "
                "will lead the backend to set the default according to "
                "'attribute_mapping_default_schema_version' option."
            ),
        )


class CreateMapping(command.ShowOne, _RulesReader):
    _description = _("Create new mapping")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'mapping',
            metavar='<name>',
            help=_('New mapping name (must be unique)'),
        )
        parser.add_argument(
            '--rules',
            metavar='<filename>',
            required=True,
            help=_('Filename that contains a set of mapping rules (required)'),
        )
        _RulesReader.add_federated_schema_version_option(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity

        rules = self._read_rules(parsed_args.rules)
        mapping = identity_client.federation.mappings.create(
            mapping_id=parsed_args.mapping,
            rules=rules,
            schema_version=parsed_args.schema_version,
        )

        mapping._info.pop('links', None)
        col_headers, col_data = zip(*sorted(mapping._info.items()))
        return col_headers, col_data


class DeleteMapping(command.Command):
    _description = _("Delete mapping(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'mapping',
            metavar='<mapping>',
            nargs='+',
            help=_('Mapping(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = self.app.client_manager.identity
        result = 0
        for i in parsed_args.mapping:
            try:
                identity_client.federation.mappings.delete(i)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete mapping with name or "
                        "ID '%(mapping)s': %(e)s"
                    ),
                    {'mapping': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.mapping)
            msg = _("%(result)s of %(total)s mappings failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListMapping(command.Lister):
    _description = _("List mappings")

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        # NOTE(marek-denis): Since rules can be long and tedious I have decided
        # to only list ids of the mappings. If somebody wants to check the
        # rules, (s)he should show specific ones.
        identity_client = self.app.client_manager.identity
        data = identity_client.federation.mappings.list()
        columns = ('ID', 'schema_version')
        items = [utils.get_item_properties(s, columns) for s in data]
        return (columns, items)


class SetMapping(command.Command, _RulesReader):
    _description = _("Set mapping properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'mapping',
            metavar='<name>',
            help=_('Mapping to modify'),
        )
        parser.add_argument(
            '--rules',
            metavar='<filename>',
            help=_('Filename that contains a new set of mapping rules'),
        )

        _RulesReader.add_federated_schema_version_option(parser)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = self.app.client_manager.identity

        rules = self._read_rules(parsed_args.rules)

        mapping = identity_client.federation.mappings.update(
            mapping=parsed_args.mapping,
            rules=rules,
            schema_version=parsed_args.schema_version,
        )

        mapping._info.pop('links', None)


class ShowMapping(command.ShowOne):
    _description = _("Display mapping details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'mapping',
            metavar='<mapping>',
            help=_('Mapping to display'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity

        mapping = identity_client.federation.mappings.get(parsed_args.mapping)

        mapping._info.pop('links', None)
        col_headers, col_data = zip(*sorted(mapping._info.items()))
        return col_headers, col_data



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.tests.functional.network.v2 import common


class NetworkFlavorProfileTests(common.NetworkTests):
    """Functional tests for network flavor profile"""

    DESCRIPTION = 'fakedescription'
    METAINFO = 'Extrainfo'

    def test_network_flavor_profile_create(self):
        json_output = self.openstack(
            'network flavor profile create '
            + '--description '
            + self.DESCRIPTION
            + ' '
            + '--enable --metainfo '
            + self.METAINFO,
            parse_output=True,
        )
        ID = json_output.get('id')
        self.assertIsNotNone(ID)
        self.assertTrue(json_output.get('enabled'))
        self.assertEqual(
            'fakedescription',
            json_output.get('description'),
        )
        self.assertEqual(
            'Extrainfo',
            json_output.get('meta_info'),
        )

        # Clean up
        raw_output = self.openstack('network flavor profile delete ' + ID)
        self.assertOutput('', raw_output)

    def test_network_flavor_profile_list(self):
        json_output = self.openstack(
            'network flavor profile create '
            + '--description '
            + self.DESCRIPTION
            + ' '
            + '--enable '
            + '--metainfo '
            + self.METAINFO,
            parse_output=True,
        )
        ID1 = json_output.get('id')
        self.assertIsNotNone(ID1)
        self.assertTrue(json_output.get('enabled'))
        self.assertEqual(
            'fakedescription',
            json_output.get('description'),
        )
        self.assertEqual(
            'Extrainfo',
            json_output.get('meta_info'),
        )

        json_output = self.openstack(
            'network flavor profile create '
            + '--description '
            + self.DESCRIPTION
            + ' '
            + '--disable '
            + '--metainfo '
            + self.METAINFO,
            parse_output=True,
        )
        ID2 = json_output.get('id')
        self.assertIsNotNone(ID2)
        self.assertFalse(json_output.get('enabled'))
        self.assertEqual(
            'fakedescription',
            json_output.get('description'),
        )
        self.assertEqual(
            'Extrainfo',
            json_output.get('meta_info'),
        )

        # Test list
        json_output = self.openstack(
            'network flavor profile list',
            parse_output=True,
        )
        self.assertIsNotNone(json_output)

        id_list = [item.get('ID') for item in json_output]
        self.assertIn(ID1, id_list)
        self.assertIn(ID2, id_list)

        # Clean up
        raw_output = self.openstack(
            'network flavor profile delete ' + ID1 + ' ' + ID2
        )
        self.assertOutput('', raw_output)

    def test_network_flavor_profile_set(self):
        json_output_1 = self.openstack(
            'network flavor profile create '
            + '--description '
            + self.DESCRIPTION
            + ' '
            + '--enable '
            + '--metainfo '
            + self.METAINFO,
            parse_output=True,
        )
        ID = json_output_1.get('id')
        self.assertIsNotNone(ID)
        self.assertTrue(json_output_1.get('enabled'))
        self.assertEqual(
            'fakedescription',
            json_output_1.get('description'),
        )
        self.assertEqual(
            'Extrainfo',
            json_output_1.get('meta_info'),
        )

        self.openstack('network flavor profile set --disable ' + ID)

        json_output = self.openstack(
            'network flavor profile show ' + ID,
            parse_output=True,
        )
        self.assertFalse(json_output.get('enabled'))
        self.assertEqual(
            'fakedescription',
            json_output.get('description'),
        )
        self.assertEqual(
            'Extrainfo',
            json_output.get('meta_info'),
        )

        # Clean up
        raw_output = self.openstack('network flavor profile delete ' + ID)
        self.assertOutput('', raw_output)

    def test_network_flavor_profile_show(self):
        json_output_1 = self.openstack(
            'network flavor profile create '
            + '--description '
            + self.DESCRIPTION
            + ' '
            + '--enable '
            + '--metainfo '
            + self.METAINFO,
            parse_output=True,
        )
        ID = json_output_1.get('id')
        self.assertIsNotNone(ID)
        json_output = self.openstack(
            'network flavor profile show ' + ID,
            parse_output=True,
        )
        self.assertEqual(
            ID,
            json_output["id"],
        )
        self.assertTrue(json_output["enabled"])
        self.assertEqual(
            'fakedescription',
            json_output["description"],
        )
        self.assertEqual(
            'Extrainfo',
            json_output["meta_info"],
        )

        # Clean up
        raw_output = self.openstack('network flavor profile delete ' + ID)
        self.assertOutput('', raw_output)



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.network.v2 import common


class NetworkQosTests(common.NetworkTests):
    def setUp(self):
        super().setUp()

        if not self.is_extension_enabled("qos"):
            self.skipTest("No qos extension present")


class NetworkQosRuleTestsMinimumBandwidth(NetworkQosTests):
    """Functional tests for QoS minimum bandwidth rule"""

    def setUp(self):
        super().setUp()

        self.QOS_POLICY_NAME = f'qos_policy_{uuid.uuid4().hex}'

        self.openstack(f'network qos policy create {self.QOS_POLICY_NAME}')
        self.addCleanup(
            self.openstack,
            f'network qos policy delete {self.QOS_POLICY_NAME}',
        )
        cmd_output = self.openstack(
            'network qos rule create '
            '--type minimum-bandwidth '
            '--min-kbps 2800 '
            f'--egress {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(
            self.openstack,
            f'network qos rule delete {self.QOS_POLICY_NAME} {self.RULE_ID}',
        )
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack(f'network qos policy create {policy_name}')
        self.addCleanup(
            self.openstack, f'network qos policy delete {policy_name}'
        )
        rule = self.openstack(
            'network qos rule create '
            '--type minimum-bandwidth '
            '--min-kbps 2800 '
            f'--egress {policy_name}',
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete {} {}'.format(policy_name, rule['id'])
        )
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            f'network qos rule list {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack(
            f'network qos rule set --min-kbps 7500 {self.QOS_POLICY_NAME} {self.RULE_ID}'
        )
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(7500, cmd_output['min_kbps'])


class NetworkQosRuleTestsMinimumPacketRate(NetworkQosTests):
    """Functional tests for QoS minimum packet rate rule"""

    def setUp(self):
        super().setUp()

        self.QOS_POLICY_NAME = f'qos_policy_{uuid.uuid4().hex}'

        self.openstack(f'network qos policy create {self.QOS_POLICY_NAME}')
        self.addCleanup(
            self.openstack,
            f'network qos policy delete {self.QOS_POLICY_NAME}',
        )
        cmd_output = self.openstack(
            'network qos rule create '
            '--type minimum-packet-rate '
            '--min-kpps 2800 '
            f'--egress {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(
            self.openstack,
            f'network qos rule delete {self.QOS_POLICY_NAME} {self.RULE_ID}',
        )
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack(f'network qos policy create {policy_name}')
        self.addCleanup(
            self.openstack, f'network qos policy delete {policy_name}'
        )
        rule = self.openstack(
            'network qos rule create '
            '--type minimum-packet-rate '
            '--min-kpps 2800 '
            f'--egress {policy_name}',
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete {} {}'.format(policy_name, rule['id'])
        )
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            f'network qos rule list {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack(
            f'network qos rule set --min-kpps 7500 --any {self.QOS_POLICY_NAME} {self.RULE_ID}'
        )
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(7500, cmd_output['min_kpps'])
        self.assertEqual('any', cmd_output['direction'])


class NetworkQosRuleTestsDSCPMarking(NetworkQosTests):
    """Functional tests for QoS DSCP marking rule"""

    def setUp(self):
        super().setUp()

        self.QOS_POLICY_NAME = f'qos_policy_{uuid.uuid4().hex}'
        self.openstack(f'network qos policy create {self.QOS_POLICY_NAME}')
        self.addCleanup(
            self.openstack,
            f'network qos policy delete {self.QOS_POLICY_NAME}',
        )
        cmd_output = self.openstack(
            'network qos rule create '
            '--type dscp-marking '
            f'--dscp-mark 8 {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(
            self.openstack,
            f'network qos rule delete {self.QOS_POLICY_NAME} {self.RULE_ID}',
        )
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack(f'network qos policy create {policy_name}')
        self.addCleanup(
            self.openstack, f'network qos policy delete {policy_name}'
        )
        rule = self.openstack(
            'network qos rule create '
            '--type dscp-marking '
            f'--dscp-mark 8 {policy_name}',
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete {} {}'.format(policy_name, rule['id'])
        )
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            f'network qos rule list {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack(
            f'network qos rule set --dscp-mark 32 {self.QOS_POLICY_NAME} {self.RULE_ID}'
        )
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(32, cmd_output['dscp_mark'])


class NetworkQosRuleTestsBandwidthLimit(NetworkQosTests):
    """Functional tests for QoS bandwidth limit rule"""

    def setUp(self):
        super().setUp()

        self.QOS_POLICY_NAME = f'qos_policy_{uuid.uuid4().hex}'
        self.openstack(f'network qos policy create {self.QOS_POLICY_NAME}')
        self.addCleanup(
            self.openstack,
            f'network qos policy delete {self.QOS_POLICY_NAME}',
        )
        cmd_output = self.openstack(
            'network qos rule create '
            '--type bandwidth-limit '
            '--max-kbps 10000 '
            f'--egress {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(
            self.openstack,
            f'network qos rule delete {self.QOS_POLICY_NAME} {self.RULE_ID}',
        )
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack(f'network qos policy create {policy_name}')
        self.addCleanup(
            self.openstack, f'network qos policy delete {policy_name}'
        )
        rule = self.openstack(
            'network qos rule create '
            '--type bandwidth-limit '
            '--max-kbps 10000 '
            '--max-burst-kbits 1400 '
            f'--egress {policy_name}',
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete {} {}'.format(policy_name, rule['id'])
        )
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            f'network qos rule list {self.QOS_POLICY_NAME}',
            parse_output=True,
        )
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack(
            'network qos rule set --max-kbps 15000 '
            '--max-burst-kbits 1800 '
            f'--ingress {self.QOS_POLICY_NAME} {self.RULE_ID}'
        )
        cmd_output = self.openstack(
            f'network qos rule show {self.QOS_POLICY_NAME} {self.RULE_ID}',
            parse_output=True,
        )
        self.assertEqual(15000, cmd_output['max_kbps'])
        self.assertEqual(1800, cmd_output['max_burst_kbps'])
        self.assertEqual('ingress', cmd_output['direction'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse

from osc_lib.cli import parseractions

from openstackclient.i18n import _


# TODO(stephenfin): Consider moving these to osc-lib since they're broadly
# useful


def add_marker_pagination_option_to_parser(
    parser: argparse.ArgumentParser,
) -> None:
    """Add marker-based pagination options to the parser.

    APIs that use marker-based paging use the marker and limit query parameters
    to paginate through items in a collection.

    Marker-based pagination is often used in cases where the length of the
    total set of items is either changing frequently, or where the total length
    might not be known upfront.
    """
    parser.add_argument(
        '--limit',
        metavar='<limit>',
        type=int,
        action=parseractions.NonNegativeAction,
        help=_(
            'The maximum number of entries to return. If the value exceeds '
            'the server-defined maximum, then the maximum value will be used.'
        ),
    )
    parser.add_argument(
        '--marker',
        metavar='<marker>',
        default=None,
        help=_(
            'The first position in the collection to return results from. '
            'This should be a value that was returned in a previous request.'
        ),
    )


def add_offset_pagination_option_to_parser(
    parser: argparse.ArgumentParser,
) -> None:
    """Add offset-based pagination options to the parser.

    APIs that use offset-based paging use the offset and limit query parameters
    to paginate through items in a collection.

    Offset-based pagination is often used where the list of items is of a fixed
    and predetermined length.
    """
    parser.add_argument(
        '--limit',
        metavar='<limit>',
        type=int,
        action=parseractions.NonNegativeAction,
        help=_(
            'The maximum number of entries to return. If the value exceeds '
            'the server-defined maximum, then the maximum value will be used.'
        ),
    )
    parser.add_argument(
        '--offset',
        metavar='<offset>',
        type=int,
        action=parseractions.NonNegativeAction,
        default=None,
        help=_(
            'The (zero-based) offset of the first item in the collection to '
            'return.'
        ),
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class RoleTests(common.IdentityTests):
    def test_role_create(self):
        self._create_dummy_role()

    def test_role_create_with_description(self):
        role_name = data_utils.rand_name('TestRole')
        description = data_utils.rand_name('description')
        raw_output = self.openstack(
            f'role create --description {description} {role_name}'
        )
        role = self.parse_show_as_object(raw_output)
        self.addCleanup(self.openstack, 'role delete {}'.format(role['id']))
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ROLE_FIELDS)
        self.assertEqual(description, role['description'])
        return role_name

    def test_role_delete(self):
        role_name = self._create_dummy_role(add_clean_up=False)
        raw_output = self.openstack(f'role delete {role_name}')
        self.assertEqual(0, len(raw_output))

    def test_role_list(self):
        self._create_dummy_role()
        raw_output = self.openstack('role list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)

    def test_role_show(self):
        role_name = self._create_dummy_role()
        raw_output = self.openstack(f'role show {role_name}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ROLE_FIELDS)

    def test_role_set(self):
        role_name = self._create_dummy_role()
        new_role_name = data_utils.rand_name('NewTestRole')
        raw_output = self.openstack(
            f'role set --name {new_role_name} {role_name}'
        )
        self.assertEqual(0, len(raw_output))
        raw_output = self.openstack(f'role show {new_role_name}')
        role = self.parse_show_as_object(raw_output)
        self.assertEqual(new_role_name, role['name'])

    def test_role_set_description(self):
        role_name = self._create_dummy_role()
        description = data_utils.rand_name("NewDescription")
        raw_output = self.openstack(
            f'role set --description {description} {role_name}'
        )
        self.assertEqual(0, len(raw_output))
        raw_output = self.openstack(f'role show {role_name}')
        role = self.parse_show_as_object(raw_output)
        self.assertEqual(description, role['description'])

    def test_role_add(self):
        role_name = self._create_dummy_role()
        username = self._create_dummy_user()
        raw_output = self.openstack(
            'role add '
            f'--project {self.PROJECT_NAME} '
            f'--project-domain {self.DOMAIN_NAME} '
            f'--user {username} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{role_name}'
        )
        self.addCleanup(
            self.openstack,
            'role remove '
            f'--project {self.PROJECT_NAME} '
            f'--project-domain {self.DOMAIN_NAME} '
            f'--user {username} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{role_name}',
        )
        self.assertEqual(0, len(raw_output))

    def test_role_add_inherited(self):
        role_name = self._create_dummy_role()
        username = self._create_dummy_user()
        raw_output = self.openstack(
            'role add '
            f'--project {self.PROJECT_NAME} '
            f'--project-domain {self.DOMAIN_NAME} '
            f'--user {username} '
            f'--user-domain {self.DOMAIN_NAME} '
            '--inherited '
            f'{role_name}'
        )
        self.addCleanup(
            self.openstack,
            'role remove '
            f'--project {self.PROJECT_NAME} '
            f'--project-domain {self.DOMAIN_NAME} '
            f'--user {username} '
            f'--user-domain {self.DOMAIN_NAME} '
            '--inherited '
            f'{role_name}',
        )
        self.assertEqual(0, len(raw_output))

    def test_role_remove(self):
        role_name = self._create_dummy_role()
        username = self._create_dummy_user()
        add_raw_output = self.openstack(
            'role add '
            f'--project {self.PROJECT_NAME} '
            f'--project-domain {self.DOMAIN_NAME} '
            f'--user {username} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{role_name}'
        )
        remove_raw_output = self.openstack(
            'role remove '
            f'--project {self.PROJECT_NAME} '
            f'--project-domain {self.DOMAIN_NAME} '
            f'--user {username} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{role_name}'
        )
        self.assertEqual(0, len(add_raw_output))
        self.assertEqual(0, len(remove_raw_output))

    def test_implied_role_list(self):
        raw_output = self.openstack('implied role list')
        default_roles = self.parse_listing(raw_output)
        self.assert_table_structure(
            default_roles, self.IMPLIED_ROLE_LIST_HEADERS
        )

        self._create_dummy_implied_role()
        raw_output = self.openstack('implied role list')
        current_roles = self.parse_listing(raw_output)
        self.assert_table_structure(
            current_roles, self.IMPLIED_ROLE_LIST_HEADERS
        )
        self.assertEqual(len(default_roles) + 1, len(current_roles))

    def test_implied_role_create(self):
        role_name = self._create_dummy_role()
        implied_role_name = self._create_dummy_role()
        self.openstack(
            'implied role create '
            f'--implied-role {implied_role_name} '
            f'{role_name}'
        )

    def test_implied_role_delete(self):
        implied_role_name, role_name = self._create_dummy_implied_role()
        raw_output = self.openstack(
            'implied role delete '
            f'--implied-role {implied_role_name} '
            f'{role_name}'
        )
        self.assertEqual(0, len(raw_output))
